body {
    padding-top: 70px;
    padding-bottom: 30px;
}


.toollist { display: table; }
.toollistrow { display: table-row; }
.tool {
    display: table-cell;
    border-width: 0;
    padding-left: 5px;
    padding-right: 5px;
    padding-top: 5px;
}

.filler {
    min-height: 2000px;
}

.affix {
    top: 50px;
    width: 100%;
    background-color: white;
}

#nav-lower {
    padding-top: 0;
}
#nav-lower > h1 {
    margin-top: 0;
}

.panel-heading a:after {
    font-family:'Glyphicons Halflings';
    content:"\e114";
    float: right;
    color: grey;
}
.panel-heading a.collapsed:after {
    content:"\e080";
}