<section id="login_form">
    <form name="login" action="javascript:sendLoginRequest();">
        <input name='server' id='server' type='text' required="required" placeholder="Server" maxlength='40' size='60' value="localhost:9000"/><br>
        <input name='user' id='user' type='text' required="required" placeholder="User" maxlength='40' size='60' /><br>
        <input name='password' id='password' required="required" placeholder="Password" type='password' maxlength='40' size='60' /><br>
        <button name="submit" type="submit">Login</button>
    </form>
</section>
