{% extends "base.html" %}
{% block content %}
</div>
<iframe style='border: none; position: absolute; background: transparent; width: 100%; height: 100%;' src="{{ url }}"></iframe>
<div class="container page">

{% endblock %}