{% extends "base.html" %}
{% block content %}

<h1>Workbench</h1>
<div class="toollist">
{% for network in networks %}
  <div class="toollistrow">
    <div class="tool">{{ network.id }}</div>
    <div class="tool">

    </div>
  </div>
{% else %}
    No networks!
{% endfor %}
</div>


{% endblock %}