{% extends "base.html" %}
{% block content %}
{% include "login.html" %}
<p>Open your browser's JavaScript console to see what's happening (hit F12).</p>
<p><input name='msg' id='msg' type="text" maxlength="512" size="100" /></p>
<p><ul>
  <li><span class="argument">FASTR#0#ECHO#"blaat"</span></li>
  <li><span class="argument">FASTR#0#CREATEWORKER#{"name": "testworker1", "user": "mkoek"}</span></li>
  <li><span class="argument">FASTR#0#CREATEWORKER#{"name": "testworker1", "user": "hachterberg"}</span></li>
  <li><span class="argument">FASTR#0#LISTWORKERS#null</span></li>
  <li><span class="argument">FASTR#0#STATUSWORKER#{"resource": FILL_WITH_WORKER_RESOURCE}</span></li>
</ul></p>
<button onclick='sendTextFromInput();'>Send Text Message from input</button>
<button onclick='sendText();'>Send Text Message</button>
<button onclick='sendBinary();'>Send Binary Message</button>
<br>
<label for="fastrLog" id="fastrLogLabel" >fastr log</label><br>
<textarea rows="15" cols="120" id="fastrLog"
        style="resize: none; height: 300px; max-height: 300px;" readonly="readonly"></textarea>

<script src="{{ url_for('static', filename='js/fastr_client.js') }}"></script>
{% endblock %}