{% extends "base.html" %}
{% block content %}


<h1>Tools Overview</h1>
<div class="toollist">
{% for tool in tools %}
  <div class="toollistrow">
    <div class="tool"><a href="{{ url_for('tool', toolname=tool[0]) }}">{{ tool[0] }}</a></div>
    <div class="tool">
    {% for version in tool[1] %}
      <a class="label label-primary" href="{{ url_for('tool', toolname=tool[0], version=version) }}">{{ version }}</a>
    {% endfor %}
    </div>
  </div>
{% else %}
  There are no tools?!
{% endfor %}
</div>

{% endblock %}