{% extends "base.html" %}
{% block content %}

<div id="nav-lower-wrapper">
  <div id="nav-lower" data-spy="affix" data-offset-top="50">
    <h1>{{ toolname }}</h1>
    <ul class="nav nav-pills" role="tablist">
      {% for version, tool in tools %}
        <li role="presentation"{% if version == selected_version %} class="active"{% endif %}>
          <a href="{{ url_for(request.endpoint, toolname=toolname, version=version) }}">{{ version }}</a>
        </li>
      {% endfor %}
    </ul>
  </div>
</div>

{% for version, tool in tools %}
  {% if selected_version == version %}
    {% include 'tool_version.html' %}
  {% endif %}
{% endfor %}

<br><br>
<div class="filler"></div>

{% endblock %}